/* API for user data dictionary */

/******************************/
/* Create a new user variable */
/******************************/
int          /* RETURNS: 1 on success, 0 on failure */
create_dd PROTO((\
   char *name,     /* Variable name */ \
   int   type,     /* See below */ \
   int   len,      /* # bytes per value */ \
   int   arrsize,  /* Max number of elements in array */ \
   int   flags,    /* See below */ \
   int   linked    /* See below */
));

/* Type should be one of: */
#define VT_STRING 0   /* String, len denotes max string length */
#define VT_INT    1   /* Integer, len is 1, 2, or 4 for byte, short, long */
#define VT_FLOAT  2   /* Float, len is 4 for float or 8 for double */
#define VT_UINT   3   /* Like VT_INT, except Unsigned integer */

/* Flags should be one of: */
#define VF_TEMPORARY  0x0001  /* not saved between runs */
#define VF_PERSISTENT 0x0003  /* saved to disk between runs */

/* Linked can be one of (or any other positive user-defined value): */
#define VL_NOTHING  0
#define VL_NODE     1  /* MBone node */
#define VL_LINK     2  /* MBone adjacency */
#define VL_CLOUD    5  /* Cloud */
#define VL_CLOUDLINK 6  /* Cloud adjacency */

/************************************************************/
/* Mark a variable as clean (unmodified since last display) */
/************************************************************/
void clean      PROTO(( char *name ));
int  is_dirty   PROTO(( char *name ));

/************************/
/* Set a variable value */
/************************/
void set_string PROTO(( char *name, int recnum, char *val));
void set_int    PROTO(( char *name, int recnum, int val));
void set_float  PROTO(( char *name, int recnum, double val));
void set_uint   PROTO(( char *name, int recnum, unsigned int val));

/* Name is the variable name.  If it doesn't exist, it's created for
 *    you with an array size of 100, len of 80/4/8, non-persistent.
 * Recnum is the record number (i.e. node #, link #, whatever) 
 * Val is the value to set 
 * It is NOT required that the value passed be of the same type as the 
 *    variable.  The value passed will be appropriately typecast for you.
 */

void set_update PROTO(( char *name, int maxage, char *updcmd, char *depends));
void set_update_interval PROTO(( char *name, int maxage ));

/* name is the variable name.
 * maxage is the number of seconds after which to update
 * updcmd is the tcl command to execute for an update.  When this is
 *    executed, the Tcl variable "exp_index" will hold the record number
 *    of the node/link/whatever whose timer expired.
 */

/* Used to delete all values associated with a node/link/whatever */
void unset PROTO((int linked, int recnum));

/* Used to delete all values associated with a variable */
void unset_all_values PROTO((char *name));

/************************/
/* Get a variable value */
/************************/
char *       get_string PROTO((char *name, int   recnum));
int          get_int    PROTO((char *name, int   recnum));
unsigned int get_uint   PROTO((char *name, int   recnum));
double       get_float  PROTO((char *name, int   recnum));

/* Name is the variable name.  If it doesn't exist, you get a value of 
 *    NULL/0/0.0 
 * Recnum is the record number (i.e. node #, link #, whatever)
 * It is NOT required that the value returned be of the same type as 
 *    the variable.  The value returned will have been
 *    appropriately typecast for you.  get_string always returns a
 *    string, thus it returns "" instead of NULL for no value.
 */

/* This is linked to the "mv_varlist <type>" Tcl command, and gives
 * a list of all user-defined variables linked to the specified type.
 *    <type> is currently one of "node", or "link"
 */
int do_varlist PROTO((ClientData data, Tcl_Interp *tcl, int argc, char **argv));

/* Linked to setting the autoupd variable in tcl */
char *set_autoupd PROTO((ClientData clientData, Tcl_Interp *interp, char *name1, char *name2, int flags));

/* Add timers should be called whenever an item is added to the display
 * so any timers associated with it will be started. 
 * Likewise, delete timers should be called whenever an item is removed 
 * from the * display.
 *    Type is the same as "linked" as detailed under create_dd() above
 *    Recnum is the record number (i.e. node #, link #, whatever) 
 * 
 * These is already done for you by display.c for all nodes and links,
 * so you probably shouldn't have to use these functions.
 */
void add_timers    PROTO((int type, int recnum));
void remove_timers PROTO((int type, int recnum));
void start_updates PROTO((char *name));
void stop_updates  PROTO((char *name));

/*
 * Get the timestamp associated with a value of an element of a
 * user variable array.  Returns timestamp, or 0 if none.
 */
time_t get_timestamp PROTO((char *name, int recnum));

/* 
 * Cause variable value to be updated according to its update command
 */
void update_value PROTO((char *name, int recnum));

int save_dd();
void save_data();
void remove_all_dd();
void remove_dd PROTO((char *name));
int get_numelts PROTO((char *name));
void dump_timers PROTO((int linked));
void combine_node_variables PROTO((int lo, int hi));
