/* $Id: dist_vect.c,v 1.16 1996/06/20 23:22:14 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/* DIST_VECT.C - computes source trees & source/dest paths       */
/*****************************************************************

  PROGRAM: Network Distance Vector Routing Computer

  DESCRIPTION:
         This program assumes knowledge of the entire
         network before starting.  It uses Dijkstra's
         algorithm to determine the shortest path tree
         from the given source and fills in the appropriate
         distance vector entries for that source in the
         sites' distance vector tables.

 METHOD:
         Dijkstra's algorithm.  The reverse direction edge weights
         are used to determine the weight between any two sites.
         The algorithm uses the relaxation method it starts with
         all nodes being infinite distance except the source which
         is zero distance.  The  graph is assumed to be stored as
         an adjacency list.  The program gradually investigates out
         from the source always expanding the lowest cost node next.
         Once a node has been investigated, it need not be looked at
         again.

  PROGRAMMER:        Deb Agarwal        6/1993
         Converted for use with mview, Dave Thaler 6/2/94

*********************************************************************/
#include <stdio.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_BSTRING_H
#include <bstring.h>
#endif
#ifdef HAVE_NETINET_IN_H
# include <sys/types.h>
# include <netinet/in.h> /* for struct in_addr for arpa/inet.h */
#endif
#include <arpa/inet.h> /* for inet_addr */
#include <tcl.h>
#include <tk.h>
#include <netdb.h>
#include <ctype.h> /* for tolower */
#include "mview.h"
#include "struct.h"
#include "node.h"
#include "display.h" /* for notify */
#include "database.h"
#include "dist_vect.h"
#include "globals.h"
#include "dict.h"
#include "link.h" /* for reverse */
#define INFINITE_DIST 32
#define INVALID_NODE  -1
#define MAX(x,y) (((x)>(y))?(x):(y))

int32    extract_min_dist_node();

typedef struct {
   int32        distance;
   int32        minttl;
   int32        hops;
   log_link_t *next_hop_subnet;
   int32        next_hop_node;
} dvect_t;
static dvect_t dist_vects[MAX_NODES];

/********* EXTERNAL REFERENCES ***********/
static int   verbose = TRUE;
static int   reverse_path = FALSE; /* if true, then use forward weights */
       int   source_node = -1, dest_node = -1;

/* Given any two nodes, find whether a link between them is on the source
 * tree
 */
int                       /* RETURNS: 1 if in tree, 0 else */
link_in_tree( src, dest ) /* ARGUMENTS: */
   int src;               /*    Source node number */
   int dest;              /*    Destination node number */
{
   if (source_node < 0)  /* No source tree exists? */
      return 0;
   return dist_vects[dest].next_hop_node == src 
       || dist_vects[src ].next_hop_node == dest;
}

/* Dump the entire path between source and destination */
void
dump_path()
{
   int v,p= -1;

   if (dest_node < 0 || source_node<0) 
      return;
   printf("\r\n");
   for (v=dest_node,p= -1; v!=source_node; p=v,v=dist_vects[v].next_hop_node) {
      printf("%d %s\r\n", v, fullname(v));
   };
   printf("%d %s\r\n", v, fullname(v));
   fflush(stdout);
}

int
do_dump_path(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   dump_path();
   return TCL_OK;
}

int
path_child(n)
   int n;
{
   int v,p= -1;

   if (dest_node<0 || source_node<0) return n;
   for (v=dest_node,p=v; v!=n && v!=source_node; p=v,v=dist_vects[v].next_hop_node);
   return (v==n)? p : n;
}

int
source_distance( n )
   int n;
{
   if (source_node < 0) return INFINITE_DIST;
   return dist_vects[n].distance;
}

int
source_parent( n )
   int n;
{
   if (source_node< 0 || dist_vects[n].next_hop_node < 0) return n;
   return dist_vects[n].next_hop_node;
}

/******************************************************************

  FUNCTION: init_source( source )

  DESCRIPTION:
        This function initializes the data structures for Dijkstra's
        algorithm which is used to create distance vector routing
        tables.

**********************************************************************/

void   
init_source( sn )
int sn;
{
   int32        v;
   log_link_t *edge;
   node_t     *node;
   int         ln;   /* link index */

   source_node = sn;
   
   for( v=0; v< num_nodes(); v++ ) {
      node = mget_node(v);
       /*
       **     Start with all nodes at infinite distance and with
       **     no parent node.
       **     The distance of the source site is 0
       */
      dist_vects[v].minttl          = (v==sn)? 0 : 255;
      dist_vects[v].distance        = (v==sn)? 0 : INFINITE_DIST;
      dist_vects[v].hops            = (v==sn)? 0 : INFINITE_DIST;
      dist_vects[v].next_hop_subnet = NULL;
      dist_vects[v].next_hop_node   = INVALID_NODE;
      node->status         &= ~ST_INTREE;

      for (ln = node->last_link[LOUT]; ln>=0; ln=llink[ln].prev_link[LOUT]) {
         edge = get_log_link(ln);

         edge->flags &= ~(LF_CHILD|LF_LEAF|LF_PARENT); /* zero child and leaf */
         edge->flags &= ~LF_INTREE;   /* used by print_tree */
      }
   }
}

/*
 * Given a dest and source, guess at the multicast reverse path between
 * them, setting the onpath_node and onpath_link variables accordingly.
 */
void
guess_path(dn,sn, val)
   int dn;
   int sn;
   int val;
{
   int v, ln;

   if (sn>=0) {
      if (dn>=0)  {
         for (v=dn; v!=sn && v>=0; v=dist_vects[v].next_hop_node) {
            set_int("onpath_node", v, val);
            ln = find_log_link(v, dist_vects[v].next_hop_node);
            if (ln>=0)
               set_int("onpath_link", ln, val);
         }
      }
      set_int("onpath_node", sn, val);
   }
}

void
guess_tree()
{
   int i, ln, dn;
   int max_nodes = num_nodes();

   for (i=0; i<max_nodes; i++) {
      if (dist_vects[i].distance < INFINITE_DIST
       && get_int("ontree_node", i) == 1) {
         set_int("ontree_node", i, 2);
         dn = dist_vects[i].next_hop_node;
         if (dn >= 0) {
            ln = find_log_link(dn, i);
            if (ln >= 0 && get_int("ontree_link", ln)==1)
               set_int("ontree_link", ln, 2);
         }
      }
   }
}

void
clear_path()
{
   int i;
   int max_nodes = num_nodes();
   int max_links = num_log_links();

   /* Set the onpath variable for all nodes & links to 'Probably False' */
   for( i = 0; i < max_nodes; i++ )
      set_int( "onpath_node", i, 1 );

   for( i = 0; i < max_links; i++ )
      set_int( "onpath_link", i, 1 );
}

void
clear_tree()
{
   int i;
   int max_nodes = num_nodes();
   int max_links = num_log_links();

   /* Set the onpath variable for all nodes & links to 'Probably False' */
   for( i = 0; i < max_nodes; i++ )
      set_int( "ontree_node", i, 1 );

   for( i = 0; i < max_links; i++ )
      set_int( "ontree_link", i, 1 );
}

void
init_dest( dn )
   int dn;
{
   int v, ln;

   /* Clear all onpath_node and onpath_link vars */
   clear_path();

   dest_node = dn;

   /* Set onpath_node and onpath_link vars */
   guess_path(dest_node, source_node, 2);
}

/**********************************************************************

  FUNCTION: relax_node( ln, src, dest, weight )

  DESCRIPTION:
       This function relaxes the dest of an edge by reassigning
       the distance function if this path is shorter than the best
       currently known path.

**********************************************************************/

void 
relax_node(rln, Redge, src, dest, weight )
   int         rln;
   log_link_t *Redge; 
   int32       src;
   int32       dest;
   int32       weight;
{
   int ln;

   /*
   **    Check if this new path is shorter distance or the same distance
   **    and through a lower address node.
   */
   if( dist_vects[dest].distance >= (dist_vects[src].distance + weight) ) {
      /*
      **      If the node this path just came from is the same
      **      length as my previous path and the address of the
      **      previous node is higher than the one I have as prev,
      **      then do not relax this node.
      */
      if( dist_vects[dest].next_hop_subnet != NULL )
         if(dist_vects[dest].distance == (dist_vects[src].distance + weight) 
          && /* Redge->to_addr */
             mget_node(Redge->dest)->iface[0]
           > /* dist_vects[dest].next_hop_subnet->to_addr */
             mget_node(dist_vects[dest].next_hop_node)->iface[0]
           )
            return;

      /*
      **    If we got this far, then this path is shorter distance than the
      **    previous known one or through a lower address so remember it as
      **    the shortest distance path.
      */
      dist_vects[dest].distance        = dist_vects[src].distance + weight;
      dist_vects[dest].hops            = dist_vects[src].hops + 1;
      ln = reverse(rln);
      if (ln>=0)
         dist_vects[dest].minttl = MAX(dist_vects[dest].hops 
          + get_int("threshold", ln), dist_vects[src].minttl);
      else /* assume symmetric */
         dist_vects[dest].minttl = MAX(dist_vects[dest].hops 
          + get_int("threshold", rln), dist_vects[src].minttl);
      dist_vects[dest].next_hop_subnet = Redge;
      dist_vects[dest].next_hop_node   = src;
   }/* endif */
   return;
}

/***********************************************************************

  FUNCTION: dijkstra_tree( source )

  DESCRIPTION:
       This function performs the outer loop of Dijkstra's algorithm.
       It initializes the node distance vector table entries, and then
       investigates nodes one at a time starting always with the one
       of shortest distance.

**********************************************************************/

int
dijkstra_tree( sn )
   int      sn;
{
   int32        cur_node, v, cnt;
   node_t     *node;
   log_link_t *edge,
              *redge;
   char        buff[255];
   int         ln;         /* link index */
   
   init_source( sn );

   while( (cur_node = extract_min_dist_node( )) != INVALID_NODE ) {
   /*
   **   cur_node is the node that is the minimum distance from the source
   **   of all the remaining nodes.  It's place in the tree has now stabilized
   **   because no shorter distance path will be found in the future so record
   **   it and relax the nodes on its outgoing edges that are not down.
   */
      node = mget_node(cur_node);
      node->status |= ST_INTREE;

      for (ln = node->last_link[LOUT]; ln>=0; ln = llink[ln].prev_link[LOUT]) {
         edge = get_log_link(ln);
         if (edge->dest < 0) continue; /* SUBNET XXX */

         if( !(edge->flags & (DVMRP_NF_DOWN | DVMRP_NF_DISABLED))) {
            int rln;

            /*
            **   The current mrouted uses the outgoing metric to determine
            **   the cost between this node and its parent in the source tree.
            **   We need to determine the reverse direction edge weight for
            **   use in the distance calculation since we are coming from the
            **   parent
            */
            rln = find_log_link( edge->dest, edge->src );
            redge = (rln>=0)? get_log_link(rln) : NULL;
            if( redge == NULL 
             || redge->flags & (DVMRP_NF_DOWN | DVMRP_NF_DISABLED)) {
           /*
           **    For some reason this tunnel was not reported in both directions
           **    so, we will assume that the tunnel is actually down
           */
               if( verbose ) {
                  sprintf(buff, "reverse edge %ld to %ld not found  or down, assuming this edge is down\n",
                   edge->dest, edge->src );
                  add_msg(buff);
               }
               edge->flags |= DVMRP_NF_DOWN;
            } else {
               if( reverse_path )
                  relax_node( rln, redge, cur_node, edge->dest, get_int("metric", ln));
               else
                  relax_node( rln, redge, cur_node, edge->dest, get_int("metric", ln));
            }
         }
      }/* endfor */
   }/* endwhile */
   /*
   **    Now that the distance vector routing calculations are complete,
   **    calculate the child and leaf bits for the reverse path forwarding
   **    algorithm.
   */
   find_child_leaf( sn ); 

   for (cnt=0,v=0; v<num_nodes(); v++)
      if (mget_node(v)->status & ST_INTREE) 
         cnt++;
   return cnt;
}

/******************************************************************************

  FUNCTION: extract_min_dist_node()

  DESCRIPTION:
        This function returns the node that has the minimum distance in
        its distance vector table.  It only searches through nodes that
        are not yet marked as in tree.  If all nodes are done, then it returns
        INVALID_NODE.  A simple linear search is performed each time the
        function is called.

******************************************************************************/

int32    
extract_min_dist_node()
{
   int32 i;
   int32 cur_min;
   int32 min_node;
   node_t *node;

   min_node = INVALID_NODE;
   cur_min  = INFINITE_DIST;
   
   for( i = 0; i < num_nodes(); i++) {
      node = mget_node(i);
      if( !(node->status & ST_INTREE) && dist_vects[i].distance < cur_min ) {
         min_node = i;
         cur_min =  dist_vects[i].distance;
      }
   }
   return min_node;
}

/***********************************************************************

  FUNCTION: find_child_leaf( source )

  DESCRIPTION:
        Fill in the child and leaf structures for each distance
        vector routing table in the graph.  It only fills in the
        entries for routes to the current source.  The child and
        leaf bits are maintained at each interface.  The calculation
        is carried out as specified for reverse path forwarding.

***********************************************************************/

void      
find_child_leaf( sn )
   int32 sn;
{
   int         v;
   int         ln; /* link # */
   log_link_t *edge;
   dvect_t    *to_dvtable;
   node_t     *snode,
              *dnode;

   for( v = 0; v < num_nodes(); v++ ) {
      snode = mget_node(v);
      if( v != sn && dist_vects[v].next_hop_subnet == NULL )
         /*
         **     This node is not in the spanning tree so we don't need
         **     to worry about its child interfaces.
         */
         continue;

      for (ln = snode->last_link[LOUT]; ln>=0; ln = llink[ln].prev_link[LOUT]) {
         edge = get_log_link(ln);
         if( !(edge->flags & (DVMRP_NF_DOWN | DVMRP_NF_DISABLED))) {
             /*
             **   If this is the next_hop_subnet, then it is the
             **   parent so mark it as such.
             */
            if( edge == dist_vects[v].next_hop_subnet ) {
               edge->flags |= LF_PARENT;
               continue;
            }
            /*
            **       This is not the parent interface so start out
            **       assuming that it is a child and leaf and then
            **       determine if it is not.
            */
            edge->flags |= ( LF_CHILD | LF_LEAF );
           /*
           **    For the neighbor (always only one) on this interface.  Check if
           **    it has a shorter distance or same distance and
           **    lower address.  If so, then this interface is not a
           **    child.
           */
            to_dvtable = &dist_vects[edge->dest];
            dnode = mget_node( edge->dest );
            if( !(dnode->status & ST_SUBNET )) {
               if( dist_vects[edge->dest].distance <= dist_vects[v].distance 
                && !(edge->flags & LF_SUBNET)) {
                  if( dist_vects[edge->dest].distance == dist_vects[v].distance ) {
                     if( dnode->iface[edge->dface] < snode->iface[edge->sface])
                        edge->flags &= ~LF_CHILD;
                  } else
                    /*
                    **     If this is a SUBNET node, then make everybody except
                    **     the parent children because messages are broadcast on
                    **     the subnet.
                    */
                     edge->flags &= ~LF_CHILD;
               }
            } else {
           /*
           **    This is a connection to a subnet.  The subnet is not a child if
           **    the subnet does not consider this node a parent.
           */
            /* if (strcmp(to_dvtable->next_hop_subnet->to_name, edge->name ))*/
               if (to_dvtable->next_hop_subnet->dest != edge->src )
                  edge->flags &= ~LF_CHILD;
            }
            if( to_dvtable->next_hop_node == INVALID_NODE 
             || to_dvtable->next_hop_node == v )
               /*
               **   This interface is not a leaf if there is a neighbor
               **   that considers this node its parent (next_hop_subnet)
               **   or if it leads to the source.
               */
               edge->flags &= ~LF_LEAF;
            
            if( (edge->flags & LF_CHILD) && (edge->flags & LF_LEAF) 
             && (edge->flags & DVMRP_NF_TUNNEL ))
               /*
               **      This interface is a tunnel and the router at the
               **      other end does not consider this router to be its
               **      parent so do not mark it as a child or leaf.
               */
               edge->flags &= ~(LF_CHILD|LF_LEAF);
         }
      }
   }
}

/* Outputs the subtree from a specific node */
void
dump_src(n, lvl)
   int n;   /* current node */
   int lvl; /* level */
{
   node_t     *node;
   log_link_t *edge;
   int i;
   int ln; /* link # */

   if (n<0) {
      printf("node<0\n");
      return;
   }

   node = mget_node(n);
   for (i=0; i<lvl; i++) { /* indent to level */
      printf("   ");
   }
   printf("%d %s\r\n", n, fullname(n));

   for (ln = node->last_link[LOUT]; ln>=0; ln = llink[ln].prev_link[LOUT]) {
      edge = get_log_link(ln);
      
      if (dist_vects[edge->dest].next_hop_node == n)
         dump_src(edge->dest, lvl+1);
   }
   fflush(stdout);
}

int
do_dump_src_tree(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   printf("\r\n");
   dump_src(source_node,0);
   return TCL_OK;
}

void
mark_source(n)
   int n;
{
   int i;
   char buff[255];

   clear_tree();
   i=dijkstra_tree( n );
   guess_tree();
   init_dest(dest_node);

   sprintf(buff, "%d", i);
   Tcl_SetResult( tcl, buff, TCL_VOLATILE );

   create_dd("distance", VT_INT, 1, MAX_NODES, VF_TEMPORARY, VL_NODE);
   create_dd("minttl",   VT_INT, 1, MAX_NODES, VF_TEMPORARY, VL_NODE);
   for (i=0; i<num_nodes(); i++) {
      set_int("distance", i, dist_vects[i].distance);
      set_int("minttl", i, dist_vects[i].minttl);
   }

   Tcl_SetVar(tcl, "srcname", fullname(n), TCL_GLOBAL_ONLY);
   Tcl_SetVar(tcl, "srcip", ipstr(n,0), TCL_GLOBAL_ONLY);

   sprintf(buff, "%s marked as source", fullname(n));
   add_msg(buff);
}

int
do_mark_source(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int n;
   char *str;
   struct hostent *hp;
   u_int32 addr;

   if (argc<2)
      n = curr_node;
   else if (isdigit(argv[1][0]))
      n = find_node_by_addr(inet_addr(argv[1]));
   else
      n = find_node_by_name(argv[1]);

   if (n<0) {
      str = yesno("Not found.  Add to database?");
      if (!str || tolower(str[0])=='n')
         n = curr_node;
      else {
         hp = gethostbyname(argv[1]);
         if (hp) {
            bcopy(hp->h_addr, (char *) &addr, hp->h_length);
            n = add_node(addr);
         } else
            notify("Error", "Unknown host");
      }
   }

   mark_source(n);
   refresh_display();
   return TCL_OK;
}
