/* $Id: event.c,v 1.7 1996/04/16 16:28:37 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
/* Tcl has something like this built in! 
 */

/* This module handles events.  An event list is kept containing the command
 * to execute, period interval, and next time of execution.  As each event's
 * time arrives, the command is executed and then rescheduled to occur after
 * the specified interval.
 *
 * Alarms can be linked to specific pieces of information about a node or
 * link.  An alarm consists of a trigger condition, rearm condition, initial
 * status, and Tcl command.  An alarm's status can be active or inactive.
 * When active, the command will be executed whenever the trigger condition
 * becomes true, and the status will change to inactive.  The status will
 * remain inactive until the rearm condition becomes true, whereupon the
 * alarm will become active again and ready to be triggered.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <sys/types.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifndef HAVE_GETTIMEOFDAY
#include <sys/timeb.h>
#endif
#include <tk.h>
#include <assert.h>
#include "mview.h"
#include "struct.h"
#include "globals.h"
#include "driver.h" /* for text_cmd */
#include "display.h" /* for do_tcl */
#include "xalloc.h" /* for xalloc */

typedef struct event_tmp {
#if (TK_MINOR_VERSION == 0)
   Tk_TimerToken token;    /* Tk's timer tag */
#else
   Tcl_TimerToken token;    /* Tk's timer tag */
#endif
#ifdef HAVE_GETTIMEOFDAY
   struct timeval time;    /* absolute time of execution                */
#else
   struct timeb time;      /* absolute time of execution                */
#endif
   int32 period;            /* period in milliseconds (0 = not periodic) */
   char command[255];      /* Tcl command to execute                    */
   struct event_tmp *next; /* pointer to next element of linked list    */
} event_t;

void process_event PROTO((event_t *this));

static event_t *event_head=NULL;

static event_t *
newEvent(sec, msec, period, command)
   int32  sec;
   int32  msec;
   int32  period;
   char *command;
{
   event_t *new;

   new = (event_t *)xalloc(0, sizeof(event_t));
#ifdef HAVE_GETTIMEOFDAY
   new->time.tv_sec  = sec;
   new->time.tv_usec = msec * 1000;
#else
   new->time.time    = sec;
   new->time.millitm = msec;
#endif
   new->period       = period;
   strcpy(new->command, command);
   new->next         = NULL;
   return new;
}

static void
deleteEvent(old)
   event_t *old;
{
   if (!old) return;
   deleteEvent(old->next);
   xfree_string(old);
}

int32
event_delta(event)
   event_t *event;
{
   int32 delta; /* milliseconds to go */
#ifdef HAVE_GETTIMEOFDAY
   struct timeval now;
   struct timezone tz;

   gettimeofday(&now,&tz);
   delta = 1000*(event->time.tv_sec  - now.tv_sec) 
              + (event->time.tv_usec - now.tv_usec)/1000;
#else
   struct timeb now;

   ftime(&now);
   delta = 1000*(event->time.time    - now.time) 
              + (event->time.millitm - now.millitm);
#endif
   return delta;
}

static void
schedule_event(curr)
   event_t *curr;
{
   event_t *next=event_head, 
           *prev=NULL;
   int32 delta;

   /* Find where it should go in the linked list */
#ifdef HAVE_GETTIMEOFDAY
   while (next && ((curr->time.tv_sec > next->time.tv_sec)
                || (curr->time.tv_sec ==next->time.tv_sec 
                 && curr->time.tv_usec> next->time.tv_usec))) 
#else
   while (next && ((curr->time.time   > next->time.time)
                || (curr->time.time   ==next->time.time 
                 && curr->time.millitm> next->time.millitm))) 
#endif
   {
        prev=next;
        next=next->next;
   }

   /* Insert it */
   curr->next = next;
   if (prev)
      prev->next = curr;
   else
      event_head = curr;

   /* Schedule it with Tk */
   delta = event_delta(curr);
   if (delta<0)  
      delta = 1; /* just to be safe, occurs on first execution */
#if (TK_MINOR_VERSION == 0)
   curr->token = Tk_CreateTimerHandler( delta, process_event, curr);
#else
   curr->token = Tcl_CreateTimerHandler( delta, process_event, curr);
#endif
}

/* This is called by Tk when the event should occur 
 * It might or might not be (?) the head event if simultaneous events
 * are scheduled.
 */
void
process_event(this)
   event_t *this;
{
   event_t *prev=NULL, 
           *curr;

   for (curr=event_head; curr && curr!=this; prev=curr,curr=curr->next);
   assert(curr);
   
   /* Remove from event list */
   if (prev)
      prev->next = curr->next;
   else
      event_head = curr->next;
   curr->next = NULL;

   /* Execute command */
   text_cmd(curr->command);

   /* Reschedule or delete event */
   if (curr->period) {
#ifdef HAVE_GETTIMEOFDAY
      int32 millitm;

      millitm = curr->period + curr->time.tv_usec/1000;
      curr->time.tv_usec = (millitm % 1000) * 1000;
      curr->time.tv_sec +=  millitm / 1000;
#else
      int32 millitm;

      millitm = curr->period + curr->time.millitm;
      curr->time.millitm = millitm % 1000;
      curr->time.time   += millitm / 1000;
#endif
      schedule_event(curr);
   } else
      deleteEvent(curr);
}

void
schedule_command(sec, msec, period, command)
   int32  sec;
   int32  msec;
   int32  period;
   char *command;
{
   schedule_event( newEvent(sec, msec, period, command) );
}

#if 0
/* removed since it was never called */
void
dump_events()
{
   event_t *curr;
   int32    delta;
   char     buff[255];

   do_tcl("open_reportwin event {Remaining    Period Command}");
   for (curr=event_head; curr; curr=curr->next) {
      delta = event_delta(curr);
      sprintf(buff, ".event.text insert end \"%9.3f %9.3f %s\"", delta/1000.0, curr->period/1000.0, 
       curr->command);
      do_tcl(buff);
   }
   do_tcl("close_reportwin event");
}
#endif
