/* $Id: node.c,v 1.26 1996/07/02 22:39:17 thalerd Exp $
 * -------------------------------------------------------------------------
 *  
 * [C] The Regents of the University of Michigan and Merit Network, Inc. 1996
 * All Rights Reserved 
 *   
 * Permission to use, copy, modify, and distribute this software and its 
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice and this permission notice appear in all 
 * copies of the software and derivative works or modified versions thereof, 
 * and that both the copyright notice and this permission and disclaimer notice 
 * appear in supporting documentation. 
 *  
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE REGENTS OF THE 
 * UNIVERSITY OF MICHIGAN AND MERIT NETWORK, INC. DO NOT WARRANT THAT THE 
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR 
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the 
 * University of Michigan and Merit Network, Inc. shall not be liable for any 
 * special, indirect, incidental or consequential damages with respect to any 
 * claim by Licensee or any third party arising from use of the software. 
 *
 * -------------------------------------------------------------------------
 */
/* NODE.C : functions operating on a single node */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/* This module is responsible for maintaining info on nodes */
#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_STRING_H
#include <string.h> /* for strcat */
#endif
#include <sys/types.h>
#include <sys/socket.h>
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#include <arpa/inet.h>
#include <time.h>
#include <tcl.h>
#include <tk.h>
#include <ctype.h>
#include "struct.h"
#include "database.h"
#include "node.h"
#include "globals.h"
#include "lib.h"
#include "display.h"
#include "maplib.h"
#include "queue.h"
#include "dict.h"
#include "xalloc.h" /* for xfree_array */
#include "driver.h" /* for text_cmd */
#include "link.h" /* for rm_link */

static int qdebug = 0;

int
do_fullname(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int n = (argc>1)? atoi(argv[1]) : curr_node;

   if (argc>1)
      Tcl_SetResult( tcl, fullname(n), TCL_STATIC );
   return TCL_OK;
}

char *      /* RETURNS: hostname of given node, or IP if blank */
fullname(n) /* ARGUMENTS: */
   int n;   /*    Node #  */
{
   char *s;

   check_node(n); /* Make sure up to date: check calls goto calls fullname */

   s = get_string("fullname", n);

   /* If no hostname, put it in the NSLOOKUP queue */
   if (!s || !s[0] || (s[0]=='?' && s[2]!='?')) {
      s=ipstr(n,0);
      push_queue(Q_NSLOOKUP, n);
      return s;
   }
   return s;
}

int
do_ipstr(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int n = (argc>1)? atoi(argv[1]) : curr_node;
   int i = (argc>2)? atoi(argv[2]) : 0;

   Tcl_SetResult( tcl, ipstr(n,i), TCL_STATIC );
   return TCL_OK;
}

char *
ipstr(n,i)
   int n;
   int i;
{
   /* check_node(n); bad news - infinite loop of calls */
   return inet_fmt(mget_node(n)->iface[i]);
}

/* Derive the hostname from the fullname in preparation for setting
 * the hostname variable.
 */
char *
hostname(n)
   int n;
{
static char hoststr[80];
   char *s, **field;

/* This doesn't work, since we're actually trying to change the value...
   s = get_string("hostname", n);
   if (s && s[0] && s[0]!='?')
      return s;
*/

   s = fullname(n);
   if (isdigit(s[0]))
      return s;
   field = explode(s,".");
   strcpy(hoststr, field[0]);
   xfree_array(field);
   return hoststr;
}

/*
 * The criteria used to determine the "domain" variable is:
 *   Second component from the right, unless it's one of: 
 *       edu,com,net,org,gov,mil,int, ac,co,go,re  (see RFC 1591)
 *    in which case use the third component from the right.
 *
 *      Thus: sydney.eecs.umich.edu          -> umich
 *            hideout-e0.sms.fi              -> sms
 *            vitruvius.arbld.unimelb.EDU.AU -> unimelb
 *            cronus.cs.ucl.ac.uk            -> ucl
 *            stockholm.mbone.ebone.net      -> ebone
 */
static char *notdomains[]={ 
   "edu","com","net","org","gov","mil","int",  /* US-style TLD's */
   "ac","co","go","re","nm",                   /* common non-US TLD's */
    NULL, };
int
valid_domain(str)
   char *str;
{
   int i;
   for (i=0; notdomains[i]; i++) {
      if (!strcasecmp(notdomains[i], str))
         return 0;
   }
   return 1;
}
char *
org(n)
   int n;
{
   char *s, **field = NULL;
static char orgstr[80];
   int nf;
   node_t *node;

   s = get_string("domain", n);
   if (s && s[0] && s[0]!='?')
      return s;

   /* Org of a subnet is org of any connected node */
   node = mget_node(n);
   if (node->status & ST_SUBNET) {
      int l = node->last_link[LOUT];
      if (l>=0)
         return org(llink[l].dest);
      else
         return get_string("domain", n);
   }

   s = fullname(n);
   if (s[0]=='?')
      strcpy(orgstr, "?");
   else {
      field = explode(s,".");
      nf = xsizeof(field);
      if (nf>2 && !valid_domain(field[nf-2]))
/*    if (nf>2 && strlen(field[ nf-1 ])==2)   */
         strcpy(orgstr, field[ nf-3 ]);
      else if (nf>1)
         strcpy(orgstr, field[ nf-2 ]);
      else
         strcpy(orgstr, s);
   }
   if (field)
      xfree_array(field);
   return orgstr;
}

char *
domain(s)
   char *s;
{
   char **field;
   int nf;
static char orgstr[80];

   field = explode(s,".");
   nf = xsizeof(field);
   if (nf>2 && strlen(field[ nf-1 ])==2)
      strcpy(orgstr, field[ nf-3 ]);
   else if (nf>1)
      strcpy(orgstr, field[ nf-2 ]);
   else
      strcpy(orgstr, s);
   xfree_array(field);
   return orgstr;
}

char *
country(n)
   int n;
{
   char *s, **field;
static char ctry[80];
   char *str;

   str = get_string("country", n);
   if (str && str[0] && str[0]!='?')
      return str;

   s = fullname(n);
   if (s[0]=='?' || isdigit(s[0]))
      strcpy(ctry, "??");
   else {
      field = explode(s,".");
      if (strlen(field[ xsizeof(field)-1 ])==2)
         strcpy(ctry, field[ xsizeof(field)-1 ]);
      else
         strcpy(ctry, "us");
      xfree_array(field);
   }
   return ctry;
}

int               /* RETURNS: 1 (yes) or 0 (no) */
needs_updating(n)
   int n;
{
   time_t  now;
   node_t *node;
   int autoupd = atoi(Tcl_GetVar(tcl, "autoupd", TCL_GLOBAL_ONLY));

   if (n<0 || !autoupd)
      return 0;

   /* Don't update if outside the area of interest */
   if (!ninfo(n))
      return 0;

   node = mget_node(n);

   if (node->status & ST_SUBNET)
      return 0;
   if (node->status & ST_DELETED)
      return 1;
   time(&now);
   return (node->timestamp < now-maxage);
}

void
remove_outlinks(n)
   int     n;
{
   node_t *node;

   if (n<0) 
      return;

   node = mget_node(n);
   while (node->last_link[LOUT]>=0)
      rm_log_link( node->last_link[LOUT] );
}

void
name_lookup(n)
   int32 n; /* Node # */
{
   char buff[256], *s;
   node_t *node;

   if (n<0)
      abort();

   node = mget_node(n);
   if (node->status & ST_SUBNET)
      return;

   if (qdebug)
      fprintf(stderr, "started nslookup of %ld\n", n);

   s = ipstr(n,0);
   sprintf(buff, "mv_start nslookup(%s) nslookup %s", s, s);
   text_cmd(buff);
}

void
autosys_lookup(n)
   int32 n; /* Node # */
{
   char buff[256], *s;

   if (n<0)
      abort();

   if (qdebug)
      fprintf(stderr, "started asn for %ld\n", n);

   s = ipstr(n,0);
   sprintf(buff, "mv_start autosys(%d) asn %s", n, s);
   text_cmd(buff);
}

void
asname_lookup(n)
   int32 n; /* Node # */
{
   char buff[256];

   if (n<0)
      abort();

   if (qdebug)
      fprintf(stderr, "started asname for %ld\n", n);

   sprintf(buff, "mv_start asname(%d) asname %d", n, get_int("autosys", n));
   text_cmd(buff);
}

void
snmpaware_lookup(n)
   int32 n; /* Node # */
{
   char buff[256], *s;
   node_t *node;

   if (n<0)
      abort();

   node = mget_node(n);
   if (node->status & ST_SUBNET) {
      pop_queue(Q_SNMPAWARE, 0); /* go on to the next one */
      return;
   }

   if (qdebug)
      fprintf(stderr, "started snmpaware for %ld\n", n);

   s = ipstr(n,0);
   sprintf(buff, "mv_start snmpaware(%d) snmpoidget -v 1 %s public .1.3.6.1.3.60.1.1.1.0 .1.3.6.1.4.1.9.10.2.1.1.1.0", n, s);
   text_cmd(buff);
}

void
update_node(n)
   int32   n; /* Node # */
{
   node_t *node;
   char mbuf[255];
   char *s1;

   if (qdebug)
      fprintf(stderr, "started update of %ld\n", n);

   if (n<0)
      abort();

   node = mget_node(n);
   if (node->status & ST_SUBNET) {
      pop_queue(Q_MRINFO, 0);
      return;
   }
   time(&(node->timestamp)); /* mark it as recent right away just for safety */

   /* Ping for up/down status */
   sprintf(mbuf, "mv_start ping /usr/etc/ping -n %s",s1=inet_fmt(node->iface[0]));
   text_cmd(mbuf);
   sprintf(mbuf, "mv_start mrinfo(%ld) mrinfo -n %s", n, s1);
   text_cmd(mbuf);

   /* Redisplay node */
   if (n==curr_node)
      goto_node(n);
}

void
check_node(n)
   int n;
{
   if (needs_updating(n)) {
      push_queue(Q_MRINFO, n);
      /*update_node(n);*/
   }
}

/*
 * Updates the status of a logical link, adding one if needed.
 * Note that if the TUNNEL flag is not set in flags, then the
 * link is actually to a subnet, and we should have links to
 * every directly connected neighbor on the subnet.
 */
int /* Returns link index */
update_link(srcaddr, dstaddr, dstmask, metric, thresh, flags)
   u_int32 srcaddr; /* Source address of node */
   u_int32 dstaddr; /* Destination address of neighbor */
   u_int32 dstmask; /* Destination subnet mask, if known (0 if not) */
   u_char  metric;
   u_char  thresh;
   u_char  flags;
{
   /*
    * Some cisco routers report tunnels to 0.0.0.0, 
    * I don't know why, but we'll count them as if they're
    * not tunnels.
    */
   if (!dstaddr)
      flags &= ~DVMRP_NF_TUNNEL;

   /* 
    * For tunnels, we just create a normal logical link 
    * between the two nodes.
    */
   if (flags & DVMRP_NF_TUNNEL) {
      int link;
      log_link_t *edge;

      link = assert_log_link(srcaddr, dstaddr, 0);
      if (link<0) 
         abort();
      edge = get_log_link(link);
      set_int("metric",    link, metric);
      set_int("threshold", link, thresh);
      edge->flags  = flags;
      time(&(edge->timestamp));

      if (flags & DVMRP_NF_DOWN)
         set_int("linkstate", link, 1); /* down */
      else {
         set_int("linkstate", link, 3); /* up */
         if (get_int("nodestate", edge->dest)<2) /* down */
             set_int("nodestate", edge->dest, 2); /* busy */

         /*
          * If no reverse edge, assume a symmetric one for starters.
          * This will be corrected if/when the remote node is updated.
          */
         if (find_log_link(edge->dest, edge->src)<0)
            update_link(dstaddr, srcaddr, 0xffffffff, metric, thresh, flags);
      }
      return link;
   } else { /* subnet */

      /* 
       * For adjacencies across a subnet, we create a subnet
       * and add links from the nodes to the subnet.
       */
      int slink, dlink, subnet;
      log_link_t *edge;

      /* Create the subnet and logical links to it */
      if (dstmask) {
         slink  = assert_log_link(srcaddr, dstaddr, dstmask);
         subnet = assert_subnet(srcaddr);
      } else {
         subnet = assert_subnet(srcaddr);
         add_iface_to_subnet(subnet, dstaddr);
         slink = assert_subnet_link(subnet, srcaddr);
         dlink = assert_subnet_link(subnet, dstaddr);

         /* Set parameters on the remote end */
         edge = get_log_link(dlink);
         set_int("metric",    dlink, metric);
         set_int("threshold", dlink, thresh);
         edge->flags  = flags;
         time(&(edge->timestamp));
   
         if (flags & DVMRP_NF_DOWN)
            set_int("linkstate", dlink, 1); /* down */
         else {
            set_int("linkstate", dlink, 3); /* up */
            if (get_int("nodestate", edge->src)<2)  /* down */
                set_int("nodestate", edge->src, 2); /* busy */
         }
      }

      /* Set the parameters on the local end */
      edge = get_log_link(slink);
      set_int("metric",    slink, metric);
      set_int("threshold", slink, thresh);
      edge->flags  = flags;
      time(&(edge->timestamp));

      if (flags & DVMRP_NF_DOWN)
         set_int("linkstate", slink, 1); /* down */
      else {
         set_int("linkstate", slink, 3); /* up */
         set_int("nodestate", edge->dest, 3); /* subnet is up */
      }

      return slink;
   }
}

/* Dump mrinfo type information */
int
dump_node(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   int n;
   node_t *node, *dnode;
   int ln=1;
   log_link_t *edge;
   char buff[255],buff2[255];

   n = (argc>1)? atoi(argv[1]) : curr_node;

   Tcl_Eval(tcl, "newwin .wtmp");
   
   wtmp_display("*******************************************************************************\n");

   node = mget_node(n);
   sprintf(buff,"%s (%s) [version %s]: %s", ipstr(n,0), 
    fullname(n), get_string("version", n), ctime(&(node->timestamp)));
   /*buff[ strlen(buff)-1 ] = '\0';*/
   wtmp_display(buff);

   for (ln = node->last_link[LOUT]; ln>=0; ln = llink[ln].prev_link[LOUT]) {
      edge = get_log_link(ln);

      sprintf(buff2,"%s -> ", ipstr(edge->src, edge->sface));
      dnode = mget_node(edge->dest);
      sprintf(buff,"%s%s (%s) [%d/%d", buff2, ipstr(edge->dest, edge->dface),
             fullname(edge->dest), get_int("metric",ln), get_int("threshold",ln));
      if (edge->flags & DVMRP_NF_TUNNEL)
         strcat(buff,"/tunnel");
      if (edge->flags & DVMRP_NF_SRCRT)
         strcat(buff,"/srcrt");
      if (edge->flags & DVMRP_NF_QUERIER)
         strcat(buff,"/querier");
      if (edge->flags & DVMRP_NF_DISABLED)
         strcat(buff,"/disabled");
      if (edge->flags & DVMRP_NF_DOWN)
         strcat(buff,"/down");
      if (edge->flags & DVMRP_NF_PIM)
         strcat(buff,"/pim");
      if (edge->flags & DVMRP_NF_LEAF)
         strcat(buff,"/leaf");
      strcat(buff,"]\n");
      /* sprintf(buff2,"%s] st=%d", buff,dnode->status);*/
      wtmp_display(buff);
   }
   wtmp_display("*******************************************************************************\n");

   return TCL_OK;
}

void
delete_node(n)
   int n;
{
   int l;

   node_t *node;
   if (n<0 || n>=num_nodes()) return;
   node =  mget_node(n);
   if (node->status & ST_DELETED)
      return;

   for (l=0; l<4; l++) {
      while (node->last_link[l]>=0)
         rm_link(node->last_link[l], n, l);
   }

   node->status |= ST_DELETED;
   delete_queue(n); /* remove from any update queues */

   /* Delete all user variables associated with this node */
   unset(VL_NODE, n);
}
