/* SPRING.C - graph layout algorithm */
/**********************************************************************/
/*                                                                    */
/*                SPRINGEMBEDDER-LAYOUT ALGORITHM                     */
/*                     based on a paper by                            */
/*              Thomas Fruchterman and Edward Reingold                */
/*                                                                    */
/*                 programmed by Claus-Dieter Ruscha                  */
/*                               at                                   */
/*                      University  of Passsau                        */
/*                                                                    */
/*               adapted for NView by Dave Thaler, 1995               */
/*                                                                    */
/**********************************************************************/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <tcl.h>
#include <tk.h>
#include <math.h>
#include <time.h>
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#include "struct.h"
#include "spring.h"
#include "node.h"
#include "database.h"
#include "display.h"
#include "globals.h"
#include "xalloc.h" /* for xfree_string */
#include "dict.h" /* for get_float */
extern int num_cloudnodes; /* imported from high.c */
extern int   cloud_ref[];

static int qdebug = 0;

typedef struct {       
   float x, y;
} vector;

typedef struct spring_edge_tmp {
   int src;
   int dest;
   int weight;
   struct spring_edge_tmp *next;
} spring_edge_t;

typedef struct {
   int    index;
   vector position;
   vector force;
   int    mark;
   spring_edge_t *source_edge;
   spring_edge_t *dest_edge;
} spring_node_t;
spring_node_t *spring_node=NULL;
int num_spring_nodes=0;

/* Check to see if the edge represents a self-loop */
static int 
reflexive_edge(edge) 
   log_link_t *edge;
{
   return (edge->src == edge->dest || edge->dest== -1);
}

/*
 * Get the weight of an edge:   1 for direct, 2 for tunnel
 */
static float
weight(ln)
   int ln;
{
   return (get_log_link(ln)->flags & DVMRP_NF_TUNNEL)? 2 : 1;
}

/**********************************************************************/
/*                                                                    */
/*             Global Variables and Data Structures                   */
/*                                                                    */
/**********************************************************************/

float   springembedder_max_force               = 3.0;
float   springembedder_opt_distance            = 64.0;
float   springembedder_vibration               = 0.4;
int   springembedder_max_iterations          = 100; /* 1000; */
int   springembedder_timelimit               = 5;

/**********************************************************************/
/*                                                                    */
/*                    P r o c e d u r e s                             */
/*                                                                    */
/**********************************************************************/

/* Clear all marked bits */
static void
unmark_all_nodes()
{    
   int i;

   for (i=0; i<num_spring_nodes; i++)
      spring_node[i].mark=0;
}

static void
mark_connected_nodes()
{
}

/* Initialize forces to all null */
static void
init_null()
{    
   int i;

   for (i=0; i<num_spring_nodes; i++) {   
      spring_node[i].force.x = 0.0;
      spring_node[i].force.y = 0.0;
   }
}

/* Compute length of vector */
static float 
vek_abs(v)
   vector v;
{     
   return hypot(v.x, v.y); /* hypot( x, y ) = sqrt( x*x + y*y ) */
}

/* Find length of longest vector */
static float 
compute_max_force()
{    
   float abs_force;
   float max_abs_force=0.0;
   int i;

   for (i=0; i<num_spring_nodes; i++) {
      abs_force = vek_abs(spring_node[i].force);
      if (abs_force > max_abs_force)
         max_abs_force = abs_force;
   }
   return max_abs_force;
}
 
static vector 
add ( v1, v2 )
   vector v1, v2;
{    
   vector v;

   v.x = ( v1.x + v2.x );
   v.y = ( v1.y + v2.y );
   return v;
}

static vector 
sub(v1,v2)
   vector v1, v2;
{
   vector v;

   v.x = ( v1.x - v2.x );
   v.y = ( v1.y - v2.y );
   return v;
}

static vector 
sca_diff(v,scalar)
   vector v;
   float scalar;
{    
   v.x = v.x / scalar;
   v.y = v.y / scalar;
   return v;
}

static void
compute_forces(optimal_distance)
   float optimal_distance;
{
   spring_node_t *node1;
   spring_edge_t *edge;
   vector attractive_force,  repulsive_force, force, delta;
   float  attractive_factor, repulsive_factor;
   float  sum_sqr_delta_xy, 
          sqr_optimal_distance = optimal_distance * optimal_distance;
   int    spring_index_1, spring_index_2;

   for (spring_index_1=0; spring_index_1 < num_spring_nodes; spring_index_1++) {
      node1 = &spring_node[spring_index_1];

      /* compute edge forces */
      for (edge = node1->source_edge; edge; edge=edge->next) {
         delta.x = spring_node[edge->dest].position.x 
                 - spring_node[edge->src ].position.x;
         delta.y = spring_node[edge->dest].position.y 
                 - spring_node[edge->src ].position.y;

         /* Prevent 0 distance, which is considered unnatural */
         /*if ((delta.x == 0) && (delta.y == 0)) */
         if ((fabs(delta.x) < 1.0) && (fabs(delta.y) < 1.0)) {  
            delta.x = 1.0;
            delta.y = 1.0;
         } 
         sum_sqr_delta_xy = delta.x * delta.x + delta.y * delta.y;

         attractive_factor  = sqrt(sum_sqr_delta_xy) 
          / ( edge->weight * optimal_distance );
         attractive_force.x = attractive_factor * delta.x;
         attractive_force.y = attractive_factor * delta.y;

         repulsive_factor  = edge->weight * edge->weight 
          * sqr_optimal_distance / sum_sqr_delta_xy;
         repulsive_force.y = repulsive_factor * delta.y;
         repulsive_force.x = repulsive_factor * delta.x;

         force.x = attractive_force.x - repulsive_force.x;
         force.y = attractive_force.y - repulsive_force.y;

         spring_node[edge->src ].force.x += force.x;
         spring_node[edge->src ].force.y += force.y;
         spring_node[edge->dest].force.x -= force.x;
         spring_node[edge->dest].force.y -= force.y;
         spring_node[edge->dest].mark = spring_index_1;
      }

      /* compute_remaining_forces */
      for (spring_index_2=spring_index_1+1; spring_index_2 < num_spring_nodes;
       spring_index_2++) {
         if ( spring_node[spring_index_2].mark != spring_index_1 ) {  
            delta.x = spring_node[spring_index_2].position.x 
                    - spring_node[spring_index_1].position.x;
            delta.y = spring_node[spring_index_2].position.y 
                    - spring_node[spring_index_1].position.y;

            /* Prevent 0 distance, which is considered unnatural */
            if ((fabs(delta.x) < 1.0) && (fabs(delta.y) < 1.0)) {  
               delta.x = 1.0;
               delta.y = 1.0;
            } 

            repulsive_factor  = sqr_optimal_distance 
             / (delta.x * delta.x + delta.y * delta.y);
            repulsive_force.x = repulsive_factor * delta.x;
            repulsive_force.y = repulsive_factor * delta.y;

            spring_node[spring_index_1].force.x -= repulsive_force.x;
            spring_node[spring_index_1].force.y -= repulsive_force.y;
            spring_node[spring_index_2].force.x += repulsive_force.x;
            spring_node[spring_index_2].force.y += repulsive_force.y;
         }
      }
   }
}
    
static void
compute_new_node_positions( temperature )
   float temperature;
{
   vector shift;
   float abs_force;
   int i;

   for ( i = 0; i < num_spring_nodes; i++ ) {   
      abs_force = vek_abs (spring_node[i].force);
      if ( abs_force < temperature) {  
         shift.x = spring_node[i].force.x;
         shift.y = spring_node[i].force.y;
      } else {  
         shift.x = spring_node[i].force.x / abs_force * temperature;
         shift.y = spring_node[i].force.y / abs_force * temperature;
      }
      spring_node[i].position.x += shift.x;
      spring_node[i].position.y += shift.y;
if (debug & DF_FORCES)
printf("%d: (%.1f, %.1f)\n", i, spring_node[i].position.x, spring_node[i].position.y);
   }
}

static int   cool_phase;
static float cool_max_force_1,
             cool_max_force_2,
             cool_max_force_3,
             cool_max_force_4;

static float 
cool(max_force)
   float max_force;
{
   static float temperature;

   cool_max_force_4 = cool_max_force_3;
   cool_max_force_3 = cool_max_force_2;
   cool_max_force_2 = cool_max_force_1;
   cool_max_force_1 = max_force;

   /* Test of Graph schwingt */
   if ( (cool_max_force_4 != 0.0) 
    && (fabs((cool_max_force_1 - cool_max_force_3) / cool_max_force_1) 
         < springembedder_vibration) 
    && (fabs((cool_max_force_2 - cool_max_force_4) / cool_max_force_2) 
         < springembedder_vibration) ) {  
      cool_phase = 2;
   }
   if ( cool_phase == 1 ) {  
      temperature = sqrt( max_force );
   } else {  /* cool_phase 2 */
      temperature = sqrt( max_force ) / 15;
   }
   return temperature;
}

/* Now rescale everything into the [100,900]x[100,900] range */
static void
rescale_spring_nodes()
{
   node_t *node;
   int i,j;
   float minx, maxx, miny, maxy;
   float xrng, yrng;
   int idx = xy_idx();

   /* If no nodes, then we're done */
   if (num_spring_nodes<1) 
      return;
     
   /* Initialize bounding box to the first node */
   minx = maxx = spring_node[0].position.x;
   miny = maxy = spring_node[0].position.y;

   /* Add all the other nodes to the bounding box */
   for (i=1; i<num_spring_nodes; i++) {
      if (spring_node[i].position.x < minx)
         minx = spring_node[i].position.x;
      if (spring_node[i].position.x > maxx)
         maxx = spring_node[i].position.x;
      if (spring_node[i].position.y < miny)
         miny = spring_node[i].position.y;
      if (spring_node[i].position.y > maxy)
         maxy = spring_node[i].position.y;
   }

   xrng = maxx-minx;
   if (!xrng)
      xrng = 10.0; /* to avoid div by 0 */
   yrng = maxy-miny;
   if (!yrng)
      yrng = 10.0;

   /* Now shift & scale all coords */
   for (i=0; i<num_spring_nodes; i++) {
      spring_node[i].position.x = 100.0 + (spring_node[i].position.x-minx) 
                                                / xrng * 800.0;
      spring_node[i].position.y = 100.0 + (spring_node[i].position.y-miny) 
                                                / yrng * 800.0;
   
   }
}

#if 0
/* Now rescale everything into the [100,900]x[100,900] range */
static void
rescale_spring_nodes()
{
   node_t *node;
   int i,j;
   float minx, maxx, miny, maxy;
   float xrng, yrng;
   int idx = xy_idx();

   if (num_spring_nodes<1) return;
     
   i = spring_node[0].index;
   if (i<0) {
      minx = maxx = cloud_x[-i];
      miny = maxy = cloud_y[-i];
   } else {
      node = mget_node(i);
      minx = maxx = node->x[idx];
      miny = maxy = node->y[idx];
   }
   
   /* Find bounding coords */
   for (j=1; j<num_spring_nodes; j++) {
      i = spring_node[j].index;
      if (i<0) {
         if (cloud_x[-i] < minx) minx = cloud_x[-i];
         if (cloud_x[-i] > maxx) maxx = cloud_x[-i];
         if (cloud_y[-i] < miny) miny = cloud_y[-i];
         if (cloud_y[-i] > maxy) maxy = cloud_y[-i];
      } else {
         node = mget_node(i);
         if (node->x[idx] < minx) minx = node->x[idx];
         if (node->x[idx] > maxx) maxx = node->x[idx];
         if (node->y[idx] < miny) miny = node->y[idx];
         if (node->y[idx] > maxy) maxy = node->y[idx];
      }
   }
   xrng = maxx-minx;
   if (!xrng)
      xrng = 10.0; /* to avoid div by 0 */
   yrng = maxy-miny;
   if (!yrng)
      yrng = 10.0;

   if (qdebug)
      printf("xrng=%f yrng=%f\n", xrng, yrng);

   /* Now shift & scale all coords */
   for (j=0; j<num_spring_nodes; j++) {
      i = spring_node[j].index;
      if (i<0) {
         cloud_x[-i] = 100.0 + (cloud_x[-i]-minx) / xrng * 800.0;
         cloud_y[-i] = 100.0 + (cloud_y[-i]-miny) / yrng * 800.0;
      } else {
         node = mget_node(i);
         node->x[idx] = 100.0 + (node->x[idx]-minx) / xrng * 800.0;
         node->y[idx] = 100.0 + (node->y[idx]-miny) / yrng * 800.0;
      }
   }
}
#endif

static void
write_node_positions_back_to_graph(cloud_idx)
   int cloud_idx; /* cloud #, or 0 for macro-topology */
{
   if (cloud_idx) { /* Micro-topology window */

      if (ndmode==ND_INCLOUD) {
         node_t     *node;
         int i,j;
     
         for (j=0; j<num_spring_nodes; j++) {
            i = spring_node[j].index;

            if (i<0) {
               cloud_x[-i] = spring_node[j].position.x;
               cloud_y[-i] = spring_node[j].position.y;
/*printf("cloud %d <= SN %d (%f,%f)\n", -i, j, cloud_x[-i], cloud_y[-i]);*/
            } else {
               node = mget_node(i);
               if (debug & DF_FORCES)
                  printf("%d: (%.1f,%.1f) changed to (%.1f,%.1f)\n", i, 
                   node->x[2], node->y[2],
                   spring_node[j].position.x,spring_node[j].position.y);
               node->x[2] = spring_node[j].position.x;
               node->y[2] = spring_node[j].position.y;
/*printf("node %d <= SN %d (%f,%f)\n", i, j, node->x[2], node->y[2]);*/
            }
         } 
      } else { /* custom */
         node_t *node;
         int j;
         int idx = xy_idx();
        
         for (j=0; j<num_spring_nodes; j++) {
            node = mget_node(spring_node[j].index);
            node->x[idx] = (int)(spring_node[j].position.x);
            node->y[idx] = (int)(spring_node[j].position.y);
   
            if (qdebug)
               printf("node %d <= SN %d (%f,%f)\n", spring_node[j].index, j, 
                node->x[idx], node->y[idx]);
         } 
      }

   } else { /* Macro-topology window */
      int i;

      for (i=0; i<num_cloudnodes; i++) {
         set_float("cloudX",i, spring_node[i].position.x );
         set_float("cloudY",i, spring_node[i].position.y );
      }
   }
}

static void
add_spring_edge(src, dest, weight)
   int   src;
   int   dest;
   float weight;
{
   spring_edge_t *sedge, *dedge;
   
   sedge = (spring_edge_t *)xalloc(0, sizeof(spring_edge_t));
   sedge->src    = src;
   sedge->dest   = dest;
   sedge->weight = weight;
   sedge->next   = spring_node[src].source_edge;
   spring_node[src].source_edge = sedge;
   
   dedge = (spring_edge_t *)xalloc(0, sizeof(spring_edge_t));
   dedge->src    = src;
   dedge->dest   = dest;
   dedge->weight = weight;
   dedge->next   = spring_node[dest].dest_edge;
   spring_node[dest].dest_edge = dedge;
}

static void
extract_cloud(cloud_idx)
   int cloud_idx; /* cloud index, or 0 for high XXX */
{
   if (cloud_idx) {
    if (ndmode==ND_INCLOUD) { /* extract cloud */
      int         cloud = cloud_idx;
      node_t     *node;
      log_link_t *edge;
      int i, n = num_nodes();
      int ln, j, k;
      int extra_clouds = 0;

      for (i=0; i<n; i++) 
         if (cloud_ref[i])
            extra_clouds++;

      num_spring_nodes = get_int("cloudsize", cloud) + extra_clouds;
      spring_node = (spring_node_t *)xalloc(0, 
                     num_spring_nodes * sizeof(spring_node_t));
      for (i=0; i<num_spring_nodes; i++) {
         spring_node[i].source_edge = NULL;
         spring_node[i].dest_edge   = NULL;
      }
      for (i=j=0; i<n && j < num_spring_nodes; i++) {
         if (get_int("cloud", i) != cloud) continue;
         node = mget_node(i);

         spring_node[j].index = i;

         if (node->x[2] > 10000.0) {
            while (node->x[2] > 1000.0)
               node->x[2] /= 10.0;
         }
         spring_node[j].position.x = node->x[2];

         if (node->y[2] > 10000.0) {
            while (node->y[2] > 1000.0)
               node->y[2] /= 10.0;
         }
         spring_node[j].position.y = node->y[2];

/*printf("SN %d <= node %d (%f,%f)\n", j, i, node->x[2], node->y[2]);*/

         spring_node[j].mark = 0; /* cloud_idx;*/
         j++;
      }

      /* Now add cloud nodes */
      for (i=0; i<n && j < num_spring_nodes; i++) {
         if (!cloud_ref[i]) continue;

         spring_node[j].index = -i;

         if (cloud_x[i] > 10000.0) {
            while (cloud_x[i] > 1000.0)
               cloud_x[i] /= 10.0;
         }
         spring_node[j].position.x = cloud_x[i];

         if (cloud_y[i] > 10000.0) {
            while (cloud_y[i] > 1000.0)
               cloud_y[i] /= 10.0;
         }
         spring_node[j].position.y = cloud_y[i];

/*printf("SN %d <= cloud %d (%f,%f)\n", j, i, cloud_x[i], cloud_y[i]);*/

         spring_node[j].mark = 0; /* cloud_idx;*/
         j++;
      }

      /* Add edges */
      for (j=0; j<num_spring_nodes; j++) {
         i = spring_node[j].index;
         node = mget_node(i);

         for (ln=node->last_link[LOUT]; ln>=0; ln=llink[ln].prev_link[LOUT]) { 
            edge = get_log_link(ln);
            if (edge->flags & LF_MULTIPLE) continue;
            /* if (get_int("cloud",edge->dest)!= cloud) continue; */
            if (edge->dest < 0 || reflexive_edge(edge)) continue;

            /* Find spring node */
            for (k=0; k<num_spring_nodes && spring_node[k].index!=edge->dest; k++);
            if (k==num_spring_nodes) {
               int cloud = get_int("cloud", edge->dest);
               for (k=0; k<num_spring_nodes && spring_node[k].index!= -cloud; k++);
            }
/*
            if (k==num_spring_nodes) {
               printf("Error in springembedder algorithm: cannot find spring node for node #%d\n", edge->dest);
               exit(1);
            }
*/

/*printf("EDGE: %d <-> %d\n", j, k);*/
            if (k<num_spring_nodes)
               add_spring_edge(j, k, weight(ln));
        }
      }

    } else { /* ndmode is CUSTOM */
      node_t     *node;
      log_link_t *edge;
      int i, n = num_nodes();
      int ln, j, k;
      int idx = xy_idx();

      spring_node = (spring_node_t *)xalloc(0, 
                     n * sizeof(spring_node_t));
      num_spring_nodes=0;
      for (i=0; i<n; i++) {
         if (!is_ndisplayed(i)) continue;
         spring_node[num_spring_nodes].source_edge = NULL;
         spring_node[num_spring_nodes].dest_edge   = NULL;
         spring_node[num_spring_nodes].index = i;

         node = mget_node(i);


         if (node->x[idx] > 10000.0) {
            while (node->x[idx] > 1000.0)
               node->x[idx] /= 10.0;
         }
         spring_node[num_spring_nodes].position.x = node->x[idx];

         if (node->y[idx] > 10000.0) {
            while (node->y[idx] > 1000.0)
               node->y[idx] /= 10.0;
         }
         spring_node[num_spring_nodes].position.y = node->y[idx];

         if (qdebug)
            printf("SN %d <= node %d (%f,%f)\n", num_spring_nodes, i, 
             node->x[idx], node->y[idx]);

         spring_node[num_spring_nodes].mark = 0; /* cloud_idx;*/

         num_spring_nodes++;
      }

      /* Add edges */
      for (j=0; j<num_spring_nodes; j++) {
         i = spring_node[j].index;
         node = mget_node(i);

         for (ln=node->last_link[LOUT]; ln>=0; ln=llink[ln].prev_link[LOUT]) { 
            if (!is_edisplayed(ln)) continue;
            edge = get_log_link(ln);
            if (edge->flags & LF_MULTIPLE) continue;
            /* if (get_int("cloud",edge->dest)!= cloud) continue; */
            if (edge->dest < 0 || reflexive_edge(edge)) continue;

            /* Find spring node */
            for (k=0; k<num_spring_nodes && spring_node[k].index!=edge->dest; k++);
            if (k==num_spring_nodes) {
               printf("Error in springembedder algorithm: cannot find spring node for node #%ld\n", edge->dest);
               exit(1);
            }

/*printf("EDGE: %d <-> %d\n", j, k);*/
            add_spring_edge(j, k, weight(ln));
        }
      }
    }
   } else {    /* extract high      */
      int i, ln;

      num_spring_nodes = num_cloudnodes;
      spring_node = (spring_node_t *)xalloc(0, 
                     num_spring_nodes * sizeof(spring_node_t));
      for (i=0; i<num_spring_nodes; i++) {
         spring_node[i].source_edge = NULL;
         spring_node[i].dest_edge   = NULL;
      }
      for (i=0; i<num_spring_nodes; i++) {
         spring_node[i].position.x = get_float("cloudX",i);
         spring_node[i].position.y = get_float("cloudY",i);
         spring_node[i].mark       = 0;

         for (ln = get_int("cloudLink", i); ln>=0; ln=get_int("cloudlinkNext", ln)) {
            add_spring_edge(get_int("cloudlinkA", ln), get_int("cloudlinkB", ln), 1.0);
         }
      }
   }
}

static void
clean_spring_edge(curr)
   spring_edge_t *curr;
{
   if (!curr) return;
   clean_spring_edge(curr->next);
   xfree_string(curr);
}

static void
clean_spring()
{
   int n;

   if (spring_node) {
      for (n=0; n<num_spring_nodes; n++) {
         clean_spring_edge(spring_node[n].source_edge);
         clean_spring_edge(spring_node[n].dest_edge);
      }

      xfree_string(spring_node);
      spring_node = NULL;
   }
   num_spring_nodes = 0;
}

/* This fixes single floating nodes */
static void
fix_floating_nodes()
{
   int i, start=1;
   float ymin=0.0, ymax=0.0, xmin=0.0, xmax=0.0;
   float xrange, yrange;

   /* Compute boundaries */
   for (i=0; i<num_spring_nodes; i++) {
      if (!spring_node[i].source_edge && !spring_node[i].dest_edge)
         continue; /* skip floating nodes */

      if (spring_node[i].position.y < ymin || start)
         ymin = spring_node[i].position.y;
      if (spring_node[i].position.x < xmin || start)
         xmin = spring_node[i].position.x;
      if (spring_node[i].position.y > ymax || start)
         ymax = spring_node[i].position.y;
      if (spring_node[i].position.x > xmax || start)
         xmax = spring_node[i].position.x;
      start=0;
   }

   /* Expand boundaries by 5% */
   xrange = xmax-xmin;
   yrange = ymax-ymin;
   xmin -= xrange/20.0;
   xmax += xrange/20.0;
   ymin -= yrange/20.0;
   ymax += yrange/20.0;
   if (!xrange) {   /* If no null boundary, create a box */
      xrange=800.0;
      xmin  =100.0;
      xmax  =900.0;
   }
   if (!yrange) {
      yrange=800.0;
      ymin  =100.0;
      ymax  =900.0;
   }

   /* Move floating nodes to boundary */
   for (i=0; i<num_spring_nodes; i++) {
      if (spring_node[i].source_edge || spring_node[i].dest_edge)
         continue; /* skip non-floating nodes */

      if (qdebug) 
         printf("Moving cloud %d\n", i);

      switch(random()%4) {
      case 0: /* top edge */
         spring_node[i].position.x = xmin + ((random()%1000) * xrange/1000.0);
         spring_node[i].position.y = ymin;
         break;

      case 1: /* bottom edge */
         spring_node[i].position.x = xmin + ((random()%1000) * xrange/1000.0);
         spring_node[i].position.y = ymax;
         break;

      case 2: /* left edge */
         spring_node[i].position.x = xmin;
         spring_node[i].position.y = ymin + ((random()%1000) * yrange/1000.0);
         break;

      case 3: /* right edge */
         spring_node[i].position.x = xmax;
         spring_node[i].position.y = ymin + ((random()%1000) * yrange/1000.0);
         break;
      }
   }
}

static void
show_stats(iter, forc)
int iter;
float forc;
{
   char buff[256];
   sprintf(buff, "I%4d Force %6.1f ", iter, forc);

   add_msg(buff);
}

void
springembedder(cloud_idx)
   int cloud_idx; /* cloud index, or 0 for high XXX */
{
   float   temperature; 
   int     iteration_count = 0;
   float   optimal_distance, 
           max_force;
   time_t  starttime;

   cool_phase       = 1;
   cool_max_force_1 = 0.0;
   cool_max_force_2 = 0.0,
   cool_max_force_3 = 0.0,
   cool_max_force_4 = 0.0;
   optimal_distance = (float)springembedder_opt_distance;

   /* Translate other data into springembedder form */
   extract_cloud(cloud_idx);

   /* Initialize forces */
   compute_forces(optimal_distance);
   max_force = compute_max_force();

   show_stats(springembedder_max_iterations-iteration_count, max_force);
   time(&starttime);
   
   /* Iterate until limit is reached on any of goodness, iterations, or time */
   while (max_force > springembedder_max_force  
    && iteration_count < springembedder_max_iterations
    && time(NULL)-starttime < springembedder_timelimit) {
      init_null();
      compute_forces(optimal_distance);
      max_force   = compute_max_force();
      temperature = cool(max_force);
      compute_new_node_positions(temperature);
      iteration_count++;
      show_stats(springembedder_max_iterations-iteration_count, max_force);
   }

   /* Fix disconnected nodes */
   fix_floating_nodes();

   /* Now rescale everything into the [0,1000]x[0,1000] range */
   rescale_spring_nodes();

   /* Save new positions */
   if (debug & DF_FORCES)
      printf("writing back\n");
   write_node_positions_back_to_graph(cloud_idx);

   /* Free up space */
   clean_spring();
}
